;; Demonstration of spreading in log time - declarations omitted for
;; clarity.  File:  spread.lisp

(in-package '*lisp)

(*DEFVAR S (self-address!!))
(*DEFVAR X (/!! (-!! (random!! (!! 10000)) (!! 5000.0)) (!! 1000.0)))
(*DEFVAR Y (!! 0.0))

;; Simple approach using pref!!  Spreads Arg values in procs m*2^k,
;; m = 0, ..., across 2^k procs, leaving result in Y (see diagram).
;;
;; |x|        |x|        |x|        |x|         | . . . |x|        |
;;  ^                    |<-------->|
;; source for              2^k procs
;; group 0

(*defun spread1 (Arg k)    ;; Arg is input pvar
 (let ((mask (- *number-of-processors-limit* (ash 1 k))))
  (*set Y (pref!! Arg (logand!! (!! mask) S))))
)

;; Potentially faster version using *pset with notification.
;; Sends along high-order address bits first.

(*defun spread2 (Arg k)
 (let* ((d (ash 1 k)) (mask (1- d)))
  (*set Y Arg)
  (*let* ((XS (self-address!!))
	  (ACTIVE (zerop!! (logand!! (!! mask) XS))))
   (dotimes (n k)        ;  d = 2^(k-n)
     (setq d (ash d -1))       ; host operation
     (*when ACTIVE
	(*pset :no-collisions Y Y (+!! XS (!! d)) :notify ACTIVE))
    )))
)

;; Open version of spread2 for demonstration

(*DEFVAR ACTV nil!!)            ; corresponds to ACTIVE
(defvar d 0) (defvar mask 0)    ; correspond to local vars
(defvar n 0)

(*defun strt2 (Arg k)  ; initialize
  (setq d (ash 1 k))  (setq mask (1- d))  (setq n 0)
  (*set Y Arg)
  (*set ACTV (zerop!! (logand!! (!! mask) S)))
)

(defun step2 ()     ; One step of spread2
  (setq d (ash d -1))
  (*when ACTV
      (*pset :no-collisions Y Y (+!! S (!! d)) :notify ACTV))
  (incf n)
)

;; Pretty-print Boolean pvar in cube order - appropriate for simulator

(*defun ppb (bv)
  (let ((stride (dimension-size 0)))
    (dotimes (j (dimension-size 1))
      (terpri)
      (dotimes (i stride)
        (write-char (if (pref bv (+ i (* j stride))) #\T #\F))
     )) ; end dotimes i, end dotimes j
 ) ; end let
 (terpri)
)